;***************************************************************
								
;     RX_Code.asm Ver. 1.0

;     Maggio 2002, Avezzano (AQ) Italia
;     Autore: Franco Anzini - email: franco.anzini@virgilio.it

;     Ricevente per telecomandi 12 bit con autoapprendimento
;     e salvataggio in eeprom del codice.

;     Usare un quarzo da 2,4576 MHz

;***************************************************************
								
		PROCESSOR 	16F84A
		RADIX		DEC
		INCLUDE		"P16F84A.INC"

;********************** Suppress MPASM warning message 302

		ERRORLEVEL	-302
		__CONFIG	3FF5H		;Wtahc dog enable

;********************** Assegnazione porte
#DEFINE		RX_Data	PORTA,0

#DEFINE		Rele_1	PORTB,0
#DEFINE		Rele_2	PORTB,1
#DEFINE		Rele_3	PORTB,2
#DEFINE		Rele_4	PORTB,3
#DEFINE		RX_Led	PORTB,4
#DEFINE		PRG_Led	PORTB,5
#DEFINE		SW_New	PORTB,6
#DEFINE		ON_OFF	PORTB,7

;********************** Routine MACRO

Wait_H		MACRO				;Macro attesa livello alto
		clrwdt				;Clear wtahc dog timer
                btfss   RX_Data
                goto    $ -0x01

		ENDM

Wait_L		MACRO				;Macro attesa livello basso
		clrwdt				;Clear wtahc dog timer
                btfsc   RX_Data
                goto    $ -0x02
		ENDM

;********************** Registri generici

                ORG     0CH

Reg_Byte_1	res	1			;Shift register codice byte 1
Reg_Byte_2	res	1			;Shift register codice byte 2

Reg_Canale	res	1			;Registro canale
Chk_Canale	res	1			;Registro controllo canale

Cod_OK_1	res	1			;Registro 1 codice in eeprom
Cod_OK_2	res	1			;Registro 2 codice in eeprom

Code_Read_1	res	1			;Registro 1 codice per confronto
Code_Read_2	res	1			;Registro 2 codice per confronto
Stato_Rele	res	1			;Registro configurazione rel

Low_Time	res	1			;Registro apprendimento start bit
Rep_Read	res	1			;Registro ripetizione letture
Conta_Bit	res	1			;Bit counter
Wait_Pulse	res	2			;Registri per ritardi attesa dati
Wait_SW		res	3			;Registri per ritardi
DelayCounter	res	3			;Registri per ritardi

;********************** Reset Vector

		ORG	0x00
Start
                bsf	STATUS,RP0		;Swap to register bank 1

		movlw	0x1F			;Set PORTA lines
                movwf	TRISA

                movlw	0xFF			;Set PORTB lines
                movwf   TRISB

                bcf     PRG_Led			;*
		bcf	RX_Led			;*
                bcf     Rele_1			;*
                bcf     Rele_2			;*** Porte OUT
                bcf     Rele_3			;*
                bcf     Rele_4			;*

		bcf	STATUS,RP0		;Swap to register bank 0

                bcf     Rele_1			;*
                bcf     Rele_2			;*
                bcf     Rele_3			;*** Disattivazione rel
                bcf     Rele_4			;*

;********************** Lettura EEPROM
Read_EE

		movlw	0x00
		movwf	EEADR
		bsf	STATUS,RP0
		bsf	EECON1,RD
		bcf	STATUS,RP0
		movf	EEDATA,W		;Lettura 1 codice
		movwf	Cod_OK_1 
		incf	EEADR,F
		bsf	STATUS,RP0
		bsf	EECON1,RD
		bcf	STATUS,RP0
		movf	EEDATA,W		;Lettura 2 codice
		movwf	Cod_OK_2 
		incf	EEADR,F
		bsf	STATUS,RP0
		bsf	EECON1,RD
		bcf	STATUS,RP0
		movf	EEDATA,W		;Lettura stato rel
		movwf	Stato_Rele 

		btfsc	ON_OFF			;Controllo ponticello
		goto	$ +0x09			;Se aperto program counter +9

		btfsc	Stato_Rele,0		;*
		bsf	Rele_1			;*
		btfsc	Stato_Rele,1		;*
		bsf	Rele_2			;*** Ripristino stato rel
		btfsc	Stato_Rele,2		;*
		bsf	Rele_3			;*
		btfsc	Stato_Rele,3		;*
		bsf	Rele_4			;*

;********************** Ripristino prescaler
ST_BY
		bsf	STATUS,RP0		;Swap to register bank 0

                movlw   0x00			;Pull-up PORTB, prescaler / 2 TMR0
                movwf   OPTION_REG

		bcf	STATUS,RP0		;Swap to register bank 0

;********************** Loop principale
Chk_RX
		clrwdt				;Clear wtahc dog timer

		bcf	PRG_Led
		bcf	RX_Led

		btfss	SW_New
		goto	Chk_SW			;Routine apprendimento codice

		btfsc	RX_Data
		goto	Rx_Code			;Routine ricezione dati

		goto	Chk_RX

;**********************	Routine ricezione dati da modulo RX
Rx_Code
		Wait_L

		movlw	0x06			;Controlla 6 volte i byte del codice
		movwf	Rep_Read

Chk_Code
		call	Wait_Code		;Attende l'inizio dei 13 bit

		movf	Reg_Byte_1,W		;Carica su W il primo byte
		subwf	Cod_OK_1,W		;Lo sottrae al primo byte presente in eeprom
		btfss	STATUS,Z		;Se errato esce
		goto	ST_BY

		movf	Reg_Byte_2,W		;Carica su W il secondo byte
		subwf   Cod_OK_2,W		;Lo sottrae al secondo byte presente in eeprom
		btfss   STATUS,Z		;Se errato esce
		goto	ST_BY

		decfsz	Rep_Read,F		;Controlla 6 volte i byte del codice
		goto	Chk_Code

;**********************	Byte codice OK! controllo canale

                movf    Reg_Canale,W		;Carica su W il byte del canale
                subwf   Chk_Canale,W		;Lo confronta con quello letto in precedenza
                btfss   STATUS,Z		;Se errato esce
                goto    ST_BY
		
		btfss	ON_OFF			;Se ponticello funzione ON/OFF inserito
		goto	Fun_ON_OFF

		movf	Reg_Canale,W		;Carica su W il byte del canale
		sublw   0x00			;Controlla se canale = 0 
		btfsc   STATUS,Z		;Se = 0 attiva rel 4
		bsf	Rele_4

		movf	Reg_Canale,W		;Carica su W il byte del canale
		sublw   0x01			;Controlla se canale = 1 
		btfsc   STATUS,Z		;Se = 1 attiva rel 1
		bsf	Rele_1

		movf	Reg_Canale,W		;Carica su W il byte del canale
		sublw   0x02			;Controlla se canale = 2 
		btfsc   STATUS,Z		;Se = 2 attiva rel 2
		bsf	Rele_2

		movf	Reg_Canale,W		;Carica su W il byte del canale
		sublw   0x03			;Controlla se canale = 3
		btfsc   STATUS,Z		;Se = 3 attiva rel 3
		bsf	Rele_3

                call    Delay			;Pausa 1"

		nop
		goto	$ -0x01			;Attiva wtahc dog

;**********************	Funzione rel bistabile 
Fun_ON_OFF
		movf	Reg_Canale,W		;Carica su W il byte del canale
		sublw   0x00			;Controlla se canale = 0
		btfsc   STATUS,Z		;Se = 0 salta al controllo rel 4
		goto	Chk_Rele_4

		movf	Reg_Canale,W		;Carica su W il byte del canale
		sublw   0x01			;Controlla se canale = 1
		btfsc   STATUS,Z		;Se = 1 salta al controllo rel 1
		goto	Chk_Rele_1

		movf	Reg_Canale,W		;Carica su W il byte del canale
		sublw   0x02			;Controlla se canale = 2
		btfsc   STATUS,Z		;Se = 2 salta al controllo rel 2
		goto	Chk_Rele_2

		movf	Reg_Canale,W		;Carica su W il byte del canale
		sublw   0x03			;Controlla se canale = 3
		btfsc   STATUS,Z		;Se = 3 salta al controllo rel 3
		goto	Chk_Rele_3

;**********************	Controllo stato rel
Chk_Rele_1
                btfss   Rele_1			;Controllo stato rel 1
                goto    Rel_1_ON

                bcf     Rele_1			;Se attivo disattiva
                bcf     Stato_Rele,0		;Clear flag rel 1
		goto	Pause
Chk_Rele_2
                btfss   Rele_2			;Controllo stato rel 2
                goto    Rel_2_ON

                bcf     Rele_2			;Se attivo disattiva
                bcf     Stato_Rele,1		;Clear flag rel 2
		goto	Pause

Chk_Rele_3
                btfss   Rele_3			;Controllo stato rel 3
                goto    Rel_3_ON

                bcf     Rele_3			;Se attivo disattiva
                bcf     Stato_Rele,2		;Clear flag rel 3
		goto	Pause

Chk_Rele_4
                btfss   Rele_4			;Controllo stato rel 4
                goto    Rel_4_ON

                bcf     Rele_4			;Se attivo disattiva
                bcf     Stato_Rele,3		;Clear flag rel 4
		goto	Pause

;**********************	Attivazione rel
Rel_1_ON
                bsf     Rele_1			;Attiva rel 1
                bsf     Stato_Rele,0		;Set flag rel 1
		goto	Pause

Rel_2_ON
                bsf     Rele_2			;Attiva rel 2
                bsf     Stato_Rele,1		;Set flag rel 2
		goto	Pause

Rel_3_ON
                bsf     Rele_3			;Attiva rel 3
                bsf     Stato_Rele,2		;Set flag rel 3
		goto	Pause

Rel_4_ON
                bsf     Rele_4			;Attiva rel 4
                bsf     Stato_Rele,3		;Set flag rel 4

;**********************	Routine salvataggio stato e pausa
Pause

		movlw	0x02			;Puntatore su terzo byte eeprom
		movwf	EEADR
		movf	Stato_Rele,W		;Salvataggio stato rel in eeprom
		call	Write_EE

                call    Delay			;Pausa 1"

		goto	ST_BY

;**********************	Routine attesa e decodifica codice
Wait_Code
		bsf	RX_Led			;Accende led ricezione

		movlw	0x03			;Attende la pausa dei dati
                movwf	Wait_Pulse+1		;per agganciare il bit di start
                clrf	Wait_Pulse+0

		clrwdt				;Clear wtahc dog timer
                btfsc   RX_Data
                goto    $ -0x06
                decfsz  Wait_Pulse+0,F
                goto    $ -0x04
                decfsz  Wait_Pulse+1,F
                goto    $ -0x06

;**********************	Misurazione del bit di start

		Wait_H				;Attende il livello alto

		clrf	TMR0			;Azzera il misuratore di durata

		clrwdt				;Clear wtahc dog timer
		movlw	0x30
		subwf	TMR0,W			;Adatta il prescaler al bit di start
		btfsc	STATUS,C		;per un range di misura ottimale
		goto	Chg_Prescaler		;Cambio prescaler
                btfsc   RX_Data
                goto    $ -0x06


		movf	TMR0,W			;Carica su W il tempo del bit di start
		addlw	0x08			;Aggiunge 8 per aumentare il margine basso
		movwf	Low_Time		;Carica il valore su Low_Time

;**********************	Decodifica primi 5 bit

		movlw	0x05
		movwf	Conta_Bit		;Legge solo i primi 5 bit

		clrf	Reg_Byte_1

		Wait_H				;Attende il livello alto

Decode_1
		clrf	TMR0			;Azzera il misuratore di durata

		bcf	Reg_Byte_1,0		;Clear bit 0

		Wait_L				;Attende il livello basso

		movf	Low_Time,W
		subwf	TMR0,W			;Controlla se bit > Low_Time
		btfsc	STATUS,C
		bsf	Reg_Byte_1,0		;Se > Low_Time set bit 0

		movf	Conta_Bit,W
		sublw	0x01			;Controlla se  l'ultimo passaggio
		btfss	STATUS,Z		;Se  l'ultimo evita lo shift del registro
		rlf	Reg_Byte_1,F

		Wait_H				;Attende il livello alto

		decfsz	Conta_Bit,F
		goto	Decode_1

;**********************	Decodifica secondi 5 bit

		movlw	0x05
		movwf	Conta_Bit		;Legge solo i secondi 5 bit

		clrf	Reg_Byte_2

Decode_2

		clrf	TMR0			;Azzera il misuratore di durata

		bcf	Reg_Byte_2,0		;Clear bit 0

		Wait_L				;Attende il livello basso

		movf	Low_Time,W
		subwf	TMR0,W			;Controlla se bit > Low_Time
		btfsc	STATUS,C
		bsf	Reg_Byte_2,0		;Se > Low_Time set bit 0

		movf	Conta_Bit,W
		sublw	0x01			;Controlla se  l'ultimo passaggio
		btfss	STATUS,Z		;Se  l'ultimo evita lo shift del registro
		rlf	Reg_Byte_2,F

		Wait_H				;Attende il livello alto

		decfsz	Conta_Bit,F
		goto	Decode_2

;**********************	Decodifica canale

		btfss	Rep_Read,2
		goto	$ +0x03			;Program counter +3
                movf    Reg_Canale,W		;Memorizza il canale letto
                movwf   Chk_Canale		;al terzo passaggio su Chk_Canale

		clrf	Reg_Canale

		clrf	TMR0			;Azzera il misuratore di durata

		bcf	Reg_Canale,0		;Clear bit 0

		Wait_L				;Attende il livello basso

		movf	Low_Time,W
		subwf	TMR0,W			;Controlla se bit > Low_Time
		btfsc	STATUS,C
		bsf	Reg_Canale,0		;Se > Low_Time set bit 0

		Wait_H				;Attende il livello alto

		clrf	TMR0			;Azzera il misuratore di durata

		bcf	Reg_Canale,1

		Wait_L				;Attende il livello basso

		movf	Low_Time,W
		subwf	TMR0,W			;Controlla se bit > Low_Time
		btfsc	STATUS,C
		bsf	Reg_Canale,1		;Se > Low_Time set bit 1

		bcf	RX_Led			;Spegne led ricezione

		return

;**********************	Routine cambio codice
Chk_SW
		movlw	0x03
		movwf	Wait_SW+2		;Imposta 3 cicli	
		clrf	Wait_SW+1		;Circa 2"
		clrf	Wait_SW+0
		clrwdt				;Clear wtahc dog timer
		btfsc	SW_New
		goto	ST_BY
		decfsz	Wait_SW+0,F
		goto	$ -0x04
		decfsz	Wait_SW+1,F
		goto	$ -0x06
		decfsz	Wait_SW+2,F
		goto	$ -0x08

New_Code
		bsf	PRG_Led			;Accende led autoapprendimento

		call	Delay			;Pausa 1"

		clrwdt				;Clear wtahc dog timer
                btfss   RX_Data			;Attende dati
		goto	$ -0x02

		movlw	0x06			;Controlla 6 volte i byte del codice
		movwf	Rep_Read

Read_Repeat

		Wait_L				;Attende il livello basso

		call	Wait_Code		;Legge i codici

                movf	Reg_Byte_1,W 
		movwf	Code_Read_1		;Carica il 1 byte letto su Code_Read_1

                movf	Reg_Byte_2,W 
		movwf	Code_Read_2		;Carica il 2 byte letto su Code_Read_2

		call	Wait_Code		;Legge i codici

		movf	Code_Read_1,W		;Carica su W Code_Read_1
		subwf	Reg_Byte_1,W		;Lo sottrae a Reg_Byte_1
		btfss	STATUS,Z		;Se errato esce
		goto	ST_BY
		
		movf	Code_Read_2,W		;Carica su W Code_Read_2
		subwf	Reg_Byte_2,W		;Lo sottrae a Reg_Byte_2
		btfss	STATUS,Z		;Se errato esce
		goto	ST_BY

		decfsz	Rep_Read,F		;6 controlli prima della scrittura in eeprom
		goto	Read_Repeat

		movlw	0x00			;Puntatore su primo byte eeprom
		movwf	EEADR
		movf	Reg_Byte_1,W		;Scrittura primo codice
		call	Write_EE

		incf	EEADR,F			;Puntatore su secondo byte eeprom
		movf	Reg_Byte_2,W		;Scrittura secondo codice
		call	Write_EE

		goto	Read_EE

;***************** Scrittura EEPROM
Write_EE
		movwf	EEDATA
		bsf	STATUS,RP0
		bsf	EECON1,WREN
		movlw	0x55
		movwf	EECON2
		movlw	0xAA
		movwf	EECON2
		bsf	EECON1,WR
Write_Code
		clrwdt				;Clear wtahc dog timer
		btfsc	EECON1,WR        
		goto	Write_Code
		bcf	STATUS,RP0

		return

;********************** Routine cambio prescaler
Chg_Prescaler
		bsf	STATUS,RP0		;Swap to register bank 0

		movf	OPTION_REG,W
		sublw   0x05
		btfsc   STATUS,Z
		goto	$ +0x04

		movf	OPTION_REG,W
                addlw	0x01			;Pull-up PORTB / prescaler 2-64 TMR0
                movwf   OPTION_REG

		bcf	STATUS,RP0		;Swap to register bank 0

		goto	Wait_Code

;********************** Routine ritardo circa 1"
Delay
                movlw   0x02
                movwf   DelayCounter+2
                clrf    DelayCounter+1
                clrf    DelayCounter+0

DelayLoop
		clrwdt				;Clear wtahc dog timer
                decfsz  DelayCounter+0,F
                goto    DelayLoop

                decfsz  DelayCounter+1,F
                goto    DelayLoop

                decfsz  DelayCounter+2,F
                goto    DelayLoop

                return

                END
